from flask import Flask, request, jsonify
from kubernetes import client, config, watch
import threading

# Load Kubernetes configuration from default location
config.load_kube_config()

# Define Kubernetes API client
k8s_client = client.ApiClient()

# Define custom resource group/version
group = 'example.com'
version = 'v1'

# Define custom resource plural
plural = 'functions'

# Define custom resource API
crd_api = client.CustomObjectsApi(k8s_client)

# Define namespace where functions will be deployed
namespace = 'default'

# Initialize Flask app
app = Flask(__name__)

# Dictionary to store function replicas
function_replicas = {}


def deploy_function(function_name, image):
    """
    Deploy a function to Kubernetes as a custom resource.
    """
    # Define function resource
    function_resource = {
        'apiVersion': f'{group}/{version}',
        'kind': 'Function',
        'metadata': {'name': function_name},
        'spec': {'image': image}
    }

    # Create function resource in Kubernetes
    k8s_client.create_namespaced_custom_object(group, version, namespace, plural, function_resource)
    print(f"Function {function_name} deployed with image {image}.")


def update_function(function_name, image):
    """
    Update an existing function in Kubernetes.
    """
    # Define function resource
    function_resource = {
        'apiVersion': f'{group}/{version}',
        'kind': 'Function',
        'metadata': {'name': function_name},
        'spec': {'image': image}
    }

    # Update function resource in Kubernetes
    k8s_client.replace_namespaced_custom_object(group, version, namespace, plural, function_name, function_resource)
    print(f"Function {function_name} updated with image {image}.")


def scale_function(function_name, replicas):
    """
    Scale the number of replicas for a function in Kubernetes.
    """
    # Define function scale spec
    scale_spec = {'spec': {'replicas': replicas}}

    # Patch function scale in Kubernetes
    k8s_client.patch_namespaced_custom_object_scale(group, version, namespace, plural, function_name, scale_spec)
    print(f"Function {function_name} scaled to {replicas} replicas.")


def handle_function(event):
    """
    Handle function events (create, update, delete).
    """
    function = event['object']
    function_name = function['metadata']['name']
    function_status = event['type']

    if function_status == 'ADDED':
        print(f"Function {function_name} added.")
        deploy_function(function_name, function['spec']['image'])
    elif function_status == 'MODIFIED':
        print(f"Function {function_name} modified.")
        update_function(function_name, function['spec']['image'])
    elif function_status == 'DELETED':
        print(f"Function {function_name} deleted.")


def watch_functions():
    """
    Watch for changes to function resources.
    """
    resource_version = ''

    while True:
        stream = watch.Watch().stream(crd_api.list_namespaced_custom_object, group, version, namespace, plural, resource_version=resource_version)
        for event in stream:
            resource_version = event['object']['metadata']['resourceVersion']
            handle_function(event)


def auto_scale():
    """
    Auto-scale functions based on request rate.
    """
    while True:
        # Example auto-scaling logic
        for function_name in function_replicas:
            # Adjust number of replicas based on request rate
            request_rate = function_replicas.get(function_name, 0)
            replicas = min(max(int(request_rate / 10), 1), 10)  # Scale between 1 and 10 replicas
            scale_function(function_name, replicas)

        # Sleep for 1 minute (adjust as needed)
        time.sleep(60)


@app.route('/invoke', methods=['POST'])
def invoke_function():
    """
    Handle HTTP POST requests to invoke functions.
    """
    data = request.get_json()
    function_name = data.get('function_name')
    # Increment request rate for function
    function_replicas[function_name] = function_replicas.get(function_name, 0) + 1
    return jsonify({'message': f"Function {function_name} invoked."})


if __name__ == '__main__':
    # Start watching for changes to function resources
    watch_thread = threading.Thread(target=watch_functions, daemon=True)
    watch_thread.start()

    # Start auto-scaling thread
    auto_scale_thread = threading.Thread(target=auto_scale, daemon=True)
    auto_scale_thread.start()

    # Run Flask app
    app.run(host='0.0.0.0', port=5000)

