import os
import subprocess
import yaml
from kubernetes import client, config

# Load Kubernetes configuration from default location
config.load_kube_config()

# Kubernetes API client
k8s_client = client.ApiClient()

# Docker Hub username (replace with your Docker Hub username)
docker_username = "msanth"

def build_and_push_image(function_name):
    """
    Build the Docker image and push it to Docker Hub.
    """
    # Change directory to function directory
    os.chdir(function_name)
    
    # Build Docker image
    subprocess.run(["docker", "build", "-t", f"{docker_username}/{function_name}:latest", "."], check=True)
    
    # Push Docker image to Docker Hub
    subprocess.run(["docker", "push", f"{docker_username}/{function_name}:latest"], check=True)
    
    # Change back to original directory
    os.chdir("..")

def create_deployment(function_name):
    """
    Create a Kubernetes Deployment resource for the function.
    """
    deployment_manifest = {
        "apiVersion": "apps/v1",
        "kind": "Deployment",
        "metadata": {"name": function_name},
        "spec": {
            "replicas": 1,
            "selector": {"matchLabels": {"app": function_name}},
            "template": {
                "metadata": {"labels": {"app": function_name}},
                "spec": {
                    "containers": [{
                        "name": function_name,
                        "image": f"{docker_username}/{function_name}:latest",
                        "ports": [{"containerPort": 8080}]  # Assuming function listens on port 8080
                    }]
                }
            }
        }
    }

    #yourusername Create Deployment in Kubernetes
    api_instance = client.AppsV1Api(k8s_client)
    api_instance.create_namespaced_deployment(namespace="default", body=deployment_manifest)
    print(f"Deployment created for {function_name}.")

def create_service(function_name):
    """
    Create a Kubernetes Service resource for the function.
    """
    service_manifest = {
        "apiVersion": "v1",
        "kind": "Service",
        "metadata": {"name": f"{function_name}-service"},
        "spec": {
            "selector": {"app": function_name},
            "ports": [{"protocol": "TCP", "port": 8080, "targetPort": 8080}],  # Assuming function listens on port 8080
            "type": "ClusterIP"  # Expose service only within the cluster
        }
    }

    # Create Service in Kubernetes
    #k8s_client.create_namespaced_service(namespace="default", body=service_manifest)
    api_instance = client.CoreV1Api()
    api_instance.create_namespaced_service(namespace="default", body=service_manifest)
    print(f"Service created for {function_name}.")

def deploy_function(function_name):
    """
    Deploy the function to Kubernetes.
    """
    build_and_push_image(function_name)
    create_deployment(function_name)
    create_service(function_name)

if __name__ == "__main__":
    function_name = input("Enter the name of the function directory: ")
    deploy_function(function_name)

