from flask import Flask, request, jsonify
from kubernetes import client, config, watch
from prometheus_client import Counter, Gauge, generate_latest, CONTENT_TYPE_LATEST
from prometheus_client.exposition import make_wsgi_app
from werkzeug.middleware.dispatcher import DispatcherMiddleware
import threading
import time
import requests
from wsgiref.simple_server import make_server

# Load Kubernetes configuration from default location
config.load_kube_config()

# Define Kubernetes API client
k8s_client = client.ApiClient()

# Initialize Flask app
app = Flask(__name__)

# Prometheus metrics
function_invocation_counter = Counter('function_invocation_count', 'Number of function invocations', ['function_name'])
function_request_rate_gauge = Gauge('function_request_rate', 'Request rate for functions', ['function_name'])


def deploy_function(function_name, image):
    """
    Deploy a function to Kubernetes as a Deployment.
    """
    # Define deployment spec
    deployment_spec = {
        'apiVersion': 'apps/v1',
        'kind': 'Deployment',
        'metadata': {'name': function_name},
        'spec': {
            'replicas': 1,
            'selector': {'matchLabels': {'app': function_name}},
            'template': {
                'metadata': {'labels': {'app': function_name}},
                'spec': {
                    'containers': [{
                        'name': function_name,
                        'image': image,
                        'ports': [{'containerPort': 8080}]
                    }]
                }
            }
        }
    }

    # Create deployment in Kubernetes
    api_instance = client.AppsV1Api(k8s_client)
    api_instance.create_namespaced_deployment(namespace='default', body=deployment_spec)
    print(f"Function {function_name} deployed with image {image}.")


def scale_function(function_name, replicas):
    """
    Scale the number of replicas for a function in Kubernetes.
    """
    # Define HPA spec
    hpa_spec = {
        'apiVersion': 'autoscaling/v1',
        'kind': 'HorizontalPodAutoscaler',
        'metadata': {'name': f'{function_name}-hpa'},
        'spec': {
            'scaleTargetRef': {
                'apiVersion': 'apps/v1',
                'kind': 'Deployment',
                'name': function_name
            },
            'minReplicas': 1,
            'maxReplicas': replicas,
            'targetCPUUtilizationPercentage': 50  # Adjust as needed
        }
    }

    # Create HPA in Kubernetes
    api_instance = client.CoreV1Api()
    k8s_client.create_namespaced_horizontal_pod_autoscaler(namespace='default', body=hpa_spec)
    print(f"Function {function_name} auto-scaler created with max replicas {replicas}.")


def handle_function(event):
    """
    Handle function events (create, update, delete).
    """
    function = event['object']
    function_name = function['metadata']['name']
    function_status = event['type']

    if function_status == 'ADDED':
        print(f"Function {function_name} added.")
        deploy_function(function_name, function['spec']['image'])
        scale_function(function_name, 10)  # Initially scale to max replicas of 10
    elif function_status == 'MODIFIED':
        print(f"Function {function_name} modified.")
        # For simplicity, we don't handle updates in this example
    elif function_status == 'DELETED':
        print(f"Function {function_name} deleted.")


def watch_functions():
    """
    Watch for changes to function resources.
    """
    resource_version = ''
    v1 = client.CoreV1Api()
    w = watch.Watch()

    while True:
        stream = w.stream(v1.list_namespaced_custom_object, group='example.com', version='v1', namespace='default', plural='functions', resource_version=resource_version)
        for event in stream:
            resource_version = event['object']['metadata']['resourceVersion']
            handle_function(event)


@app.route('/invoke', methods=['POST'])
def invoke_function():
    """
    Handle HTTP POST requests to invoke functions.
    """
    data = request.get_json()
    function_name = data.get('function_name')
    # Increment request rate for function
    function_request_rate_gauge.labels(function_name=function_name).inc()
    # Increment invocation count for function
    function_invocation_counter.labels(function_name=function_name).inc()
    # Invoke function
    result = invoke_kubernetes_function(function_name)
    return jsonify({'message': f"Function {function_name} invoked. Result: {result}"})


def invoke_kubernetes_function(function_name):
    """
    Invoke a Kubernetes function by sending an HTTP request.
    """
    # Assuming the function is exposed via a service with type ClusterIP
    # You need to replace 'function-service' with the actual service name in your Kubernetes cluster
    url = f"http://function-service.default.svc.cluster.local:8080/invoke"

    # Example payload, adjust as needed
    payload = {
        "message": "Hello from Python!"
    }

    try:
        response = requests.post(url, json=payload)
        if response.status_code == 200:
            return response.json()  # Assuming the function returns JSON response
        else:
            return f"Error: {response.status_code}"
    except requests.exceptions.RequestException as e:
        return f"Error: {e}"

def metrics_app(environ, start_response):
    """
    Serve Prometheus metrics.
    """
    response_body = generate_latest()
    status = '200 OK'
    headers = [('Content-type', CONTENT_TYPE_LATEST)]
    start_response(status, headers)
    return [response_body]


if __name__ == '__main__':
    # Start watching for changes to function resources
    watch_thread = threading.Thread(target=watch_functions, daemon=True)
    watch_thread.start()

    # Run Flask app
    app.run(host='0.0.0.0', port=5000)

    # Serve Prometheus metrics
    metrics_app_wrapper = DispatcherMiddleware(app.wsgi_app, {'/metrics': metrics_app})
    httpd = make_server('0.0.0.0', 8080, metrics_app_wrapper)
    httpd.serve_forever()

